import unittest
from datetime import datetime, timedelta, timezone
from typing import Optional
from unittest import skip

import pandas as pd

from chalk.df.ChalkDataFrameImpl import ChalkDataFrameImpl
from chalk.features import DataFrame, Filter, after, feature_time, features, has_many, has_one
from chalk.features.feature import Filter
from chalk.features.resolver import ResolverAnnotationParser


@features
class ToppingPrice:
    id: str
    topping_id: str
    wow: str
    ts: datetime = feature_time()


@features
class Nickname:
    id: str


@features
class Taco:
    id: str
    user_id: str
    price: int
    maybe_price: Optional[int]
    hat: str
    topping_id: str
    ts: datetime = feature_time()
    nicknames: DataFrame[Nickname] = has_many(lambda: Nickname.id == Taco.id)
    topping: "Topping" = has_one(lambda: Topping.id == Taco.topping_id)


@features
class Topping:
    id: str
    ts: datetime = feature_time()
    nicknames: DataFrame[Nickname] = has_many(lambda: Nickname.id == Topping.id)
    price: ToppingPrice = has_one(lambda: ToppingPrice.topping_id == Topping.id)


class TestDF(unittest.TestCase):
    @skip
    def test_indexing(self):
        # TODO: Make this work
        f = DataFrame.from_dict({Taco.id: [1, 2, 3], Taco.price: [5, 6, 7]})
        self.assertEqual(Taco(id=3, price=7), f[-1])
        self.assertEqual(Taco(id=2, price=6), f[-1])
        self.assertEqual(Taco(id=1, price=5), f[-1])

    def test_range(self):
        f = DataFrame.from_dict({Taco.id: [1, 2, 3]})
        self.assertTrue(f[0:].count() == f.count() == 3)
        self.assertTrue(f[1:].count() == 2)
        self.assertTrue(f[0:1].count() == 1)

    def test_eq(self):
        self.assertEqual(
            DataFrame.from_dict({Taco.id: [1, 1, 3]}),
            DataFrame.from_dict({Taco.id: [1, 1, 3]}),
        )
        self.assertNotEqual(
            DataFrame.from_dict({Taco.id: [1, 1]}),
            DataFrame.from_dict({Taco.id: [1, 1, 3]}),
        )

    def test_mode(self):
        f = DataFrame.from_dict({Taco.id: [1, 1, 3]})
        self.assertTrue(f.mode() == 1)
        self.assertTrue(f[Taco.id != 1].mode() == 3)

    def test_empty(self):
        non_empty = DataFrame.from_dict({Taco.id: [1]})
        self.assertTrue(non_empty.is_not_empty)
        self.assertFalse(non_empty.is_empty)

        for empty in [
            DataFrame.from_dict({Taco.id: []}),
            DataFrame.from_dict({}),
        ]:
            self.assertTrue(empty.is_empty)
            self.assertFalse(empty.is_not_empty)

    def test_in(self):
        chalk_df = DataFrame.from_dict(
            {
                Taco.id: ["t_1", "t_2", "t_3"],
                Taco.price: [1, 2, 10],
            }
        )
        one_two_three_tacos = chalk_df[Taco.price in (1, 2, 3)]
        assert isinstance(one_two_three_tacos, ChalkDataFrameImpl)
        assert one_two_three_tacos.count() == 2
        assert one_two_three_tacos[Taco.price].max() == 2
        assert one_two_three_tacos[Taco.price].min() == 1
        self.assertEqual(chalk_df[Taco.price in []].count(), 0)
        self.assertEqual(chalk_df[Taco.price in [444]].count(), 0)
        self.assertEqual(chalk_df[Taco.price not in [1, 4]].count(), 2)
        self.assertEqual(chalk_df[not Taco.price in [1]].count(), 2)

    def test_is_none(self):
        chalk_df = DataFrame.from_dict(
            {
                Taco.id: ["t_1", "t_2", "t_3"],
                Taco.maybe_price: [20, None, 10],
            }
        )
        self.assertEqual(chalk_df[Taco.maybe_price is None].count(), 1)
        self.assertEqual(chalk_df[Taco.maybe_price == None].count(), 1)
        self.assertEqual(chalk_df[Taco.maybe_price is not None].count(), 2)
        self.assertEqual(chalk_df[Taco.maybe_price != None].count(), 2)

    def test_bool_op(self):
        chalk_df = DataFrame.from_dict(
            {
                Taco.id: ["t_1", "t_2", "t_3"],
                Taco.maybe_price: [20, None, 10],
                Taco.price: [1, 2, 3],
            }
        )
        self.assertEqual(chalk_df[Taco.maybe_price is None or Taco.id == "t_1"].count(), 2)
        self.assertEqual(chalk_df[Taco.maybe_price is None and Taco.id == "t_2"].count(), 1)
        self.assertEqual(
            chalk_df[not (Taco.maybe_price is not None or not Taco.id == "t_2")].count(),
            1,
        )

        self.assertEqual(chalk_df[Taco.maybe_price is None and Taco.id == "t_1"].count(), 0)
        self.assertEqual(
            chalk_df[Taco.maybe_price is not None and (Taco.id == "t_1" or Taco.price == 3)].count(),
            2,
        )
        self.assertEqual(
            chalk_df[Taco.maybe_price is not None and (Taco.id == "t_1" or Taco.price in [8, 4, 99])].count(),
            1,
        )

    def test_after(self):
        now = datetime.now()
        chalk_df = DataFrame.from_dict(
            {
                Taco.ts: pd.Series([now - timedelta(days=10), now], dtype="datetime64[ns, utc]"),
                Taco.price: [5, 10],
            }
        )
        new_price = chalk_df[after(days_ago=5)].sum()
        self.assertEqual(new_price, 10)

    def test_filtering(self):
        chalk_df = DataFrame.from_dict(
            {
                Taco.id: ["t_1", "t_2", "t_3"],
                Taco.price: [1, 5, 10],
            }
        )
        expensive_tacos = chalk_df[Taco.price > 1]
        self.assertIsInstance(expensive_tacos, ChalkDataFrameImpl)
        self.assertEqual(expensive_tacos.count(), 2)
        with self.assertRaises(ValueError) as e:
            expensive_tacos.max()
        self.assertEqual(
            str(e.exception),
            (
                "Cannot compute max. DataFrame contains 2 features, and expected only one. Filter your "
                "DataFrame down to "
                "the feature for which you want to compute the max before calling .max()."
            ),
        )

        self.assertEqual(expensive_tacos[Taco.price].max(), 10)
        self.assertEqual(expensive_tacos[Taco.price].min(), 5)

    def test_filter_and_project(self):
        chalk_df = DataFrame.from_dict(
            {
                Taco.user_id: ["u_1", "u_1", "u_2"],
                Taco.id: ["t_1", "t_2", "t_3"],
                Taco.price: [1, 5, 10],
            }
        )
        expensive_tacos = chalk_df[Taco.price, Taco.price > 1, Taco.user_id == "u_1"]
        self.assertIsInstance(expensive_tacos, ChalkDataFrameImpl)
        self.assertEqual(expensive_tacos.count(), 1)
        self.assertEqual(expensive_tacos.max(), 5)
        self.assertEqual(expensive_tacos.min(), 5)

    def test_filter_many(self):
        chalk_df = DataFrame.from_dict(
            {
                Taco.user_id: ["u_1", "u_1", "u_2"],
                Taco.id: ["t_1", "t_2", "t_3"],
                Taco.price: [1, 5, 10],
            }
        )
        expensive_tacos = chalk_df[Taco.price > 1, Taco.user_id == "u_1"]
        self.assertIsInstance(expensive_tacos, ChalkDataFrameImpl)
        self.assertEqual(expensive_tacos.count(), 1)
        with self.assertRaises(ValueError) as e:
            expensive_tacos.max()
        self.assertEqual(
            str(e.exception),
            (
                "Cannot compute max. DataFrame contains 3 features, and expected only one. Filter your "
                "DataFrame down to "
                "the feature for which you want to compute the max before calling .max()."
            ),
        )

        self.assertEqual(expensive_tacos[Taco.price].max(), 5)
        self.assertEqual(expensive_tacos[Taco.price].min(), 5)

    def test_filtering_with_datetime(self):
        chalk_df = DataFrame.from_dict(
            {
                Taco.user_id: ["u_1", "u_1", "u_2"],
                Taco.id: ["t_1", "t_2", "t_3"],
                Taco.price: [1, 5, 10],
                Taco.ts: [
                    datetime.now(tz=timezone.utc) - timedelta(days=3),
                    datetime.now(tz=timezone.utc) - timedelta(days=1),
                    datetime.now(tz=timezone.utc),
                ],
            }
        )
        recent_tacos = chalk_df[Taco.ts >= datetime.now(tz=timezone.utc) - timedelta(hours=3)]
        self.assertIsInstance(recent_tacos, ChalkDataFrameImpl)
        self.assertEqual(recent_tacos.count(), 1)
        self.assertEqual(recent_tacos[Taco.price].max(), 10)

    def test_indexing_one(self):
        chalk_df = DataFrame.from_dict(
            {
                Taco.id: ["t_1", "t_2", "t_3"],
                Taco.price: [100, 200, 300],
            }
        )
        self.assertEqual(chalk_df[Taco.price].mean(), 200)
        self.assertEqual(chalk_df[Taco.price].count(), 3)
        self.assertEqual(chalk_df[Taco.price].max(), 300)
        self.assertEqual(chalk_df[Taco.price].min(), 100)

    def test_indexing_many(self):
        chalk_df = DataFrame.from_dict(
            {
                Taco.id: ["t_1", "t_2", "t_3"],
                Taco.price: [100, 200, 300],
                Taco.hat: ["green", "red", "blue"],
            }
        )
        self.assertTrue(
            chalk_df[Taco.price, Taco.id].underlying.equals(
                pd.DataFrame(
                    {
                        Taco.price: [100, 200, 300],
                        Taco.id: ["t_1", "t_2", "t_3"],
                    }
                )
            )
        )

    def test_input_feature_types(self):
        df = DataFrame.from_dict({Taco.topping.id: ["1"]})
        self.assertEqual(df[Taco.topping.id].iloc[0, 0], "1")

    def test_feature_to_string(self):
        def _check(string: str, feat):
            self.assertEqual(string, str(feat))
            self.assertEqual(string, str(feat))

        # scalar feature
        _check("taco.price", Taco.price)

        # ts feature
        _check("taco.ts", Taco.ts)

        # has-one feature
        _check("taco.topping", Taco.topping)

        # scalar feature has-one
        _check("taco.topping.id", Taco.topping.id)

        # ts feature has-one
        _check("taco.topping.ts", Taco.topping.ts)

        _check("taco.topping.nicknames", Taco.topping.nicknames)

        _check("taco.topping.price.ts", Taco.topping.price.ts)

        # Double has-one
        _check("taco.topping.price.wow", Taco.topping.price.wow)

        _check("taco.nicknames", Taco.nicknames)

    def test_from_dict(self):
        df = DataFrame.from_dict(
            {
                Taco.topping.id: [100, 200],
                Taco.id: [1, 3],
            }
        )
        self.assertEqual(300, df[Taco.topping.id].sum())
        self.assertEqual(4, df[Taco.id].sum())

    def test_to_pandas(self):
        df = DataFrame.from_dict(
            {
                Taco.topping.id: [100, 200],
            }
        )
        pds = df.to_pandas()
        self.assertEqual(300, pds[Taco.topping.id].sum())

    def test_from_list(self):
        self.assertEqual(0, DataFrame.from_list().count())
        self.assertEqual(0, DataFrame.from_list([]).count())
        self.assertEqual(1, DataFrame.from_list(Taco(price=123, user_id="33")).count())
        self.assertEqual(1, DataFrame.from_list([Taco(price=123, user_id="33")]).count())
        self.assertEqual(
            2,
            DataFrame.from_list(Taco(price=123, user_id="33"), Taco(price=3, user_id="33")).count(),
        )
        self.assertEqual(
            2,
            DataFrame.from_list([Taco(price=123, user_id="33"), Taco(price=3, user_id="33")]).count(),
        )
        with self.assertRaises(ValueError):
            DataFrame.from_list(Taco(price=123, user_id="33"), Taco(price=3))

        with self.assertRaises(ValueError):
            DataFrame.from_list(Taco(), Taco(price=3, user_id="33"))

    def test_ast_parsing(self):
        container = {1, 2, 3}

        def my_resolver(
            my_arg: DataFrame[
                Taco.price,
                Taco.user_id > "33",
                Taco.maybe_price is not None,
                Taco.maybe_price is None,
                Taco.price not in container,
                Taco.price in container or (Taco.topping.price == -3 and (Taco.hat == "cool" or Taco.price == 4)),
            ]
        ):
            ...

        fancy_index = ResolverAnnotationParser(my_resolver, globals(), locals()).parse_annotation("my_arg")
        assert issubclass(fancy_index, DataFrame)

        self.assertEqual(">", fancy_index.filters[0].operation)
        self.assertEqual("taco.user_id", str(fancy_index.filters[0].lhs))
        self.assertEqual("33", fancy_index.filters[0].rhs)

        self.assertEqual("!=", fancy_index.filters[1].operation)
        self.assertEqual("taco.maybe_price", str(fancy_index.filters[1].lhs))
        self.assertEqual(None, fancy_index.filters[1].rhs)

        self.assertEqual("==", fancy_index.filters[2].operation)
        self.assertEqual("taco.maybe_price", str(fancy_index.filters[2].lhs))
        self.assertEqual(None, fancy_index.filters[2].rhs)

        self.assertEqual("not in", fancy_index.filters[3].operation)
        self.assertEqual("taco.price", str(fancy_index.filters[3].lhs))
        self.assertEqual(container, fancy_index.filters[3].rhs)

        fancy_clause = fancy_index.filters[4]
        assert isinstance(fancy_clause, Filter)
        first_fancy = fancy_clause.lhs
        assert isinstance(first_fancy, Filter)
        self.assertEqual("in", first_fancy.operation)
        self.assertEqual("taco.price", str(first_fancy.lhs))
        self.assertEqual(container, first_fancy.rhs)

        self.assertEqual("or", fancy_clause.operation)



import collections.abc
import datetime
import enum
import functools
import logging
from typing import Any, List, Sequence, Union

import pandas as pd

from chalk.df.ast_parser import parse_dataframe_getitem
from chalk.features import Feature
from chalk.features.feature import FeatureWrapper, Filter, TimeDelta, unwrap_feature
from chalk.utils.collections import ensure_tuple
from chalk.utils.duration import convert_datetime_to_pd

_logger = logging.getLogger(__name__)


def _feature_type_or_value(e: Union[Feature, FeatureWrapper]):
    if isinstance(e, FeatureWrapper):
        e = unwrap_feature(e)
    return e


class ChalkDataFrameImpl:
    def __init__(self, pandas_dataframe: pd.DataFrame):
        self.underlying = pandas_dataframe
        self.timestamp_feature = None
        # All columns in a ChalkDataFrameImpl must be UNPARSED features, since they interact with client code
        # Converting all column labels into Feature instances so column labels match how users
        # would index into the dataframe -- for example df[User.account.number]
        def _to_feature_with_path(col: Any):
            if isinstance(col, FeatureWrapper):
                col = unwrap_feature(col)
            if isinstance(col, Feature):
                return col
            if isinstance(col, str):
                return Feature.from_root_fqn(col)
            raise TypeError(
                f"Column {col} has unexpected type: {type(col).__name__}. Column types must be Feature instances or strings of Root FQNs."
            )

        pandas_dataframe.columns = pandas_dataframe.columns.map(_to_feature_with_path)
        for col in pandas_dataframe.columns:
            assert isinstance(col, Feature)
            if col.is_feature_time:
                self.timestamp_feature = col

    def _expect_scalar(self, v, name: str):
        if isinstance(v, pd.Series) and len(v) == 1:
            return v[0]
        if isinstance(v, pd.Series):
            raise ValueError(
                (
                    f"Cannot compute {name}. DataFrame contains {len(v)} features, and expected only one. "
                    f"Filter your DataFrame down to the feature for which you want to compute the {name} "
                    f"before calling .{name}()."
                )
            )
        return v

    def mean(self):
        return self._expect_scalar(self.underlying.mean(), "mean")

    def median(self):
        return self._expect_scalar(self.underlying.median(), "median")

    def mode(self):
        return self._expect_scalar(self.underlying.median(), "mode")

    def __eq__(self, other: object):
        if isinstance(other, ChalkDataFrameImpl):
            return other.underlying.equals(self.underlying)
        if isinstance(other, pd.DataFrame):
            return self.underlying.equals(other)
        return NotImplemented

    def sum(self):
        return self._expect_scalar(self.underlying.sum(), "sum")

    def max(self):
        return self._expect_scalar(self.underlying.max(), "max")

    def min(self):
        return self._expect_scalar(self.underlying.min(), "min")

    def count(self) -> int:
        return len(self.underlying)

    def __len__(self) -> int:
        return len(self.underlying)

    @property
    def is_empty(self) -> bool:
        return self.underlying.empty

    @property
    def is_not_empty(self) -> bool:
        return not self.is_empty

    def keys(self):
        return self.underlying.keys()

    def iterrows(self):
        return self.underlying.iterrows()

    @property
    def iloc(self):
        return self.underlying.iloc()

    def __repr__(self):
        return repr(self.underlying)

    def _maybe_replace_timestamp_feature(self, f: Union[Feature, Any]):
        """Replace the ``CHALK_TS`` pseudo-feature with the actual timestamp column."""
        if not isinstance(f, Feature) or f.fqn != "__chalk__.CHALK_TS":
            return f
        if self.timestamp_feature is None:
            raise ValueError("DataFrame has no timestamp")
        return self.timestamp_feature

    def _maybe_convert_timedelta_to_timestamp(
        self, f: Union[TimeDelta, datetime.timedelta, Any], now: datetime.datetime
    ):
        """Convert timedeltas relative to ``now`` into absolute datetimes."""
        if isinstance(f, TimeDelta):
            f = f.to_std()
        if isinstance(f, datetime.timedelta):
            return convert_datetime_to_pd(now + f)
        return f

    def _parse_feature_or_value(self, f: Union[Feature, Any], now: datetime.datetime):
        """Parse a feature or value into the correct type that can be used for filtering."""
        f = _feature_type_or_value(f)
        f = self._maybe_convert_timedelta_to_timestamp(f, now)
        f = self._maybe_replace_timestamp_feature(f)
        if isinstance(f, enum.Enum):
            f = f.value
        return f

    def apply_filters(self, filters: Sequence[Filter], now: datetime.datetime) -> "ChalkDataFrameImpl":
        """Apply multiple filters on the dataframe.

        The filters will be joined together with AND.

        Args:
            filters: The filters
            now: The datetime to use for the current timestamp. Used to resolve relative
                timestamps in filters to absolute datetimes.

        Returns:
            A new dataframe with all filters applied.
        """
        underlying = self.underlying

        if len(filters) > 0:
            pandas_filters = [self._evaluate_filter(f, now) for f in filters]
            single_filter = functools.reduce(lambda a, b: a & b, pandas_filters)
            underlying = underlying[single_filter]

        return ChalkDataFrameImpl(underlying)

    def _evaluate_filter(self, f: Filter, now: datetime.datetime) -> pd.Series:
        """Evaluate a filter.

        Args:
            f (Filter): The filter.
            now (datetime.datetime): The datetime to use for the current timestamp. Used to resolve relative
                timestamps in filters to absolute datetimes.

        Returns:
            A series of boolean values that can be used to select the rows where the filter is truthy
        """
        # Passing `now` in explicitly instead of using datetime.datetime.now() so that multiple filters
        # relying on relative timestamps (e.g. before, after) will have the same "now" time.
        if f.operation == "not":
            assert f.rhs is None, "not has just one side"
            assert isinstance(f.lhs, Filter), "lhs must be a filter"
            return ~self._evaluate_filter(f.lhs, now)
        elif f.operation == "and":
            assert isinstance(f.rhs, Filter), "rhs must be a filter"
            assert isinstance(f.lhs, Filter), "lhs must be a filter"
            return self._evaluate_filter(f.lhs, now) & self._evaluate_filter(f.rhs, now)
        elif f.operation == "or":
            assert isinstance(f.rhs, Filter), "rhs must be a filter"
            assert isinstance(f.lhs, Filter), "lhs must be a filter"
            return self._evaluate_filter(f.lhs, now) | self._evaluate_filter(f.rhs, now)

        lhs = self._parse_feature_or_value(f.lhs, now)
        rhs = self._parse_feature_or_value(f.rhs, now)

        if isinstance(lhs, Feature):
            lhs = self.underlying[lhs]

        if isinstance(rhs, Feature):
            rhs = self.underlying[rhs]

        if rhs is None:
            assert isinstance(lhs, pd.Series)
            if f.operation == "==":
                return lhs.isnull()

            elif f.operation == "!=":
                return lhs.notnull()

        if isinstance(lhs, pd.Series) and lhs.dtype.name != "object" and isinstance(rhs, enum.Enum):
            rhs = rhs.value

        if isinstance(rhs, pd.Series) and rhs.dtype.name != "object" and isinstance(lhs, enum.Enum):
            lhs = lhs.value

        if f.operation == "==":
            ret = lhs == rhs
        elif f.operation == "!=":
            ret = lhs != rhs
        elif f.operation == ">=":
            ret = lhs >= rhs
        elif f.operation == ">":
            ret = lhs > rhs
        elif f.operation == "<":
            ret = lhs < rhs
        elif f.operation == "<=":
            ret = lhs <= rhs
        elif f.operation in ("in", "not in"):
            assert isinstance(lhs, pd.Series)
            assert isinstance(rhs, collections.abc.Iterable)
            new_rhs = []
            for x in rhs:
                if lhs.dtype.name != "object" and isinstance(x, enum.Enum):
                    new_rhs.append(x.value)
                else:
                    new_rhs.append(x)
            ret = lhs.isin(new_rhs)
            if f.operation == "not in":
                ret = ~ret
        else:
            raise ValueError(f'Unknown operation "{f.operation}"')
        assert isinstance(ret, pd.Series)
        return ret

    def __getitem__(self, item):
        has_bool_or_filter_value = any(isinstance(x, (bool, Filter)) for x in ensure_tuple(item))
        if has_bool_or_filter_value:
            # If we have a boolean or Filter value, then that means we need to ast-parse the caller since
            # python has already evaluated AND, OR, and IN operations into literal booleans or Filters
            # Skipping the parsing unless if we have need to for effeciency and to eliminate conflicts
            # with pytest
            item = parse_dataframe_getitem()
        try:
            projections: list[Feature] = []
            filters: List[Filter] = []
            slices = []
            for x in ensure_tuple(item):
                if isinstance(x, FeatureWrapper):
                    x = unwrap_feature(x)
                if isinstance(x, Feature):
                    projections.append(x)

                elif isinstance(x, Filter):
                    filters.append(x)

                elif isinstance(x, slice):
                    slices.append(x)

                elif isinstance(x, int):
                    raise ValueError("Indexing into the chalk.DataFrame is not yet supported")

                else:
                    raise ValueError(f'Indexed chalk.DataFrame with invalid type "{type(x).__name__}": {x}')

            now = datetime.datetime.now(tz=datetime.timezone.utc)
            self_filtered = self.apply_filters(filters, now)
            underlying = self_filtered.underlying
            if len(slices) > 0:
                underlying = underlying.iloc[tuple(slices)]

            # Do the projection
            underlying = underlying[projections] if len(projections) > 0 else underlying
            assert isinstance(underlying, pd.DataFrame)

            return ChalkDataFrameImpl(underlying)

        except KeyError:
            _logger.debug(f"Failed to get key: Had keys: {self.underlying.keys()}", exc_info=True)
            raise

    def to_pandas(self) -> Union[pd.DataFrame, pd.Series]:
        return self.underlying
